from git_autograder.answers.answers_record import GitAutograderAnswersRecord
from git_autograder.answers.rules.answer_rule import AnswerRule


class NotEmptyRule(AnswerRule):
    def apply(self, answer: GitAutograderAnswersRecord) -> None:
        if answer.answer.strip() != "":
            raise Exception(f"Answer for {answer.question} is empty.")



import functools
import os
from datetime import datetime
from typing import Callable

import pytz

from git_autograder.exception import (
    GitAutograderInvalidStateException,
    GitAutograderWrongAnswerException,
)
from git_autograder.output import GitAutograderOutput
from git_autograder.repo import GitAutograderRepo
from git_autograder.status import GitAutograderStatus


def autograder() -> Callable[
    [Callable[..., GitAutograderOutput]], Callable[..., GitAutograderOutput]
]:
    """
    Decorator to denote that a function is an autograder function.

    Initializes the GitAutograderRepo and provides it as an argument to the function.

    Handles the Git Autograder exceptions thrown by the GitAutograderRepo or from wrong
    answers.

    All outputs are saved directly to disk.
    """

    def inner(
        func: Callable[..., GitAutograderOutput],
    ) -> Callable[..., GitAutograderOutput]:
        @functools.wraps(func)
        def wrapper(*args, **kwargs) -> GitAutograderOutput:
            output = None
            is_local = os.environ.get("is_local", "false") == "true"

            exercise_name = os.environ.get("exercise_name")
            if exercise_name is None:
                output = GitAutograderOutput(
                    exercise_name=None,
                    started_at=None,
                    completed_at=None,
                    is_local=is_local,
                    comments=["Missing exercise_name in environment"],
                    status=GitAutograderStatus.ERROR,
                )
                output.save()
                return output

            repo = GitAutograderRepo(is_local=is_local, exercise_name=exercise_name)
            try:
                output = func(repo, *args, **kwargs)
            except (
                GitAutograderInvalidStateException,
                GitAutograderWrongAnswerException,
            ) as e:
                output = GitAutograderOutput(
                    exercise_name=repo.exercise_name,
                    started_at=repo.started_at,
                    completed_at=datetime.now(tz=pytz.UTC),
                    is_local=repo.is_local,
                    comments=[e.message] if isinstance(e.message, str) else e.message,
                    status=(
                        GitAutograderStatus.ERROR
                        if isinstance(e, GitAutograderInvalidStateException)
                        else GitAutograderStatus.UNSUCCESSFUL
                    ),
                )
            except Exception as e:
                # Unexpected exception
                output = GitAutograderOutput(
                    exercise_name=exercise_name,
                    started_at=None,
                    completed_at=None,
                    is_local=is_local,
                    comments=[str(e)],
                    status=GitAutograderStatus.ERROR,
                )

            assert output is not None
            output.save()
            return output

        return wrapper

    return inner



from typing import List

from git import Commit, Repo

from git_autograder.exception import GitAutograderInvalidStateException
from git_autograder.repo import GitAutograderRepo


class CommitHelper:
    def __init__(self, repo: Repo) -> None:
        self.repo = repo

    def is_child_commit(self, child: Commit, parent: Commit) -> bool:
        if child == parent:
            return True

        res = False
        for parent in child.parents:
            res |= self.is_child_commit(parent, parent)

        return res

    def commits(self, branch: str = "main") -> List[Commit]:
        """Retrieve the available commits of a given branch."""
        commits = []
        for commit in self.repo.iter_commits(branch):
            commits.append(commit)

        return commits

    def start_commit(self, branch: str = "main") -> Commit:
        """
        Find the Git Mastery start commit from the given branch.

        Raises exceptions if the branch has no commits or if the start tag is not
        present.
        """
        commits = self.commits(branch)

        if len(commits) == 0:
            raise GitAutograderInvalidStateException(
                f"Branch {branch} is missing any commits",
            )

        first_commit = commits[-1]

        first_commit_hash = first_commit.hexsha
        start_tag_name = f"git-mastery-start-{first_commit_hash[:7]}"

        start_tag = None
        for tag in self.repo.tags:
            if str(tag) == start_tag_name:
                start_tag = tag
                break

        if start_tag is None:
            raise GitAutograderInvalidStateException(
                f"Branch {branch} is missing the Git Mastery start commit",
            )

        return start_tag.commit

    def user_commits(self, branch: str = "main") -> List[Commit]:
        """
        Retrieves only the user commits from a given branch.

        Raises exceptions if the branch has no commits or start tag is not present.
        """
        start_commit = self.start_commit(branch)
        commits = self.commits(branch)
        commits_asc = list(reversed(commits))
        start_commit_index = commits_asc.index(start_commit)
        user_commits = commits_asc[start_commit_index + 1 :]

        return user_commits



from typing import List

from git import Repo

from git_autograder.exception import GitAutograderInvalidStateException


class BranchHelper:
    def __init__(self, repo: Repo) -> None:
        self.repo = repo

    def track_remote_branches(self, remotes: List[str], strict: bool = False) -> None:
        if "origin" not in [remote.name for remote in self.repo.remotes]:
            return

        tracked = {"main"}
        for remote in self.repo.remote("origin").refs:
            for r in remotes:
                if r not in tracked or f"origin/{r}" != remote.name:
                    continue
                tracked.add(r)
                self.repo.git.checkout("-b", r, f"origin/{r}")
                break

        missed_remotes = list(set(remotes).difference(tracked))
        if len(missed_remotes) > 0 and strict:
            raise GitAutograderInvalidStateException(
                f"Missing branches {', '.join(missed_remotes)} in submission",
            )

    def has_branch(self, branch: str) -> bool:
        return branch in self.repo.heads



import json
import os
from dataclasses import asdict, dataclass
from datetime import datetime
from typing import ClassVar, List, Optional

from git_autograder.encoder import Encoder
from git_autograder.status import GitAutograderStatus


@dataclass
class GitAutograderOutput:
    status: GitAutograderStatus

    started_at: Optional[datetime]
    completed_at: Optional[datetime]
    is_local: Optional[bool]
    comments: Optional[List[str]] = None
    exercise_name: Optional[str] = None

    OUTPUT_FILE_NAME: ClassVar[str] = "output.json"

    def save(self, path: str = "../output") -> None:
        os.makedirs(path, exist_ok=True)
        file_path = os.path.join(path, self.OUTPUT_FILE_NAME)
        output = asdict(self)
        with open(file_path, "w") as f:
            f.write(json.dumps(output, cls=Encoder))



from dataclasses import dataclass
from typing import Iterator, Optional

from difflib_parser.difflib_parser import DiffParser
from git import Commit, DiffIndex
from git.diff import Diff, Lit_change_type


@dataclass
class GitAutograderDiff:
    diff: Diff
    change_type: Lit_change_type
    original_file_path: Optional[str]
    edited_file_path: Optional[str]
    original_file: Optional[str]
    edited_file: Optional[str]
    diff_parser: Optional[DiffParser]


class GitAutograderDiffHelper:
    def __init__(self, a: Commit, b: Commit) -> None:
        self.diff_index: DiffIndex[Diff] = a.diff(b)

    def iter_changes(self, change_type: Lit_change_type) -> Iterator[GitAutograderDiff]:
        for change in self.diff_index.iter_change_type(change_type):
            original_file_rawpath = change.a_rawpath
            edited_file_rawpath = change.b_rawpath
            original_file_path = (
                original_file_rawpath.decode("utf-8")
                if original_file_rawpath is not None
                else None
            )
            edited_file_path = (
                edited_file_rawpath.decode("utf-8")
                if edited_file_rawpath is not None
                else None
            )
            original_file_blob = change.a_blob
            edited_file_blob = change.b_blob
            original_file = (
                original_file_blob.data_stream.read().decode("utf-8")
                if original_file_blob is not None
                else None
            )
            edited_file = (
                edited_file_blob.data_stream.read().decode("utf-8")
                if edited_file_blob is not None
                else None
            )

            diff_parser = (
                DiffParser(original_file.split("\n"), edited_file.split("\n"))
                if original_file is not None and edited_file is not None
                else None
            )

            yield GitAutograderDiff(
                change_type=change_type,
                diff=change,
                original_file_path=original_file_path,
                edited_file_path=edited_file_path,
                original_file=original_file,
                edited_file=edited_file,
                diff_parser=diff_parser,
            )



from dataclasses import dataclass
from typing import List, Optional

from git_autograder.answers.answers_record import GitAutograderAnswersRecord


@dataclass
class GitAutograderAnswers:
    questions: List[str]
    answers: List[str]

    @property
    def qna(self) -> List[GitAutograderAnswersRecord]:
        return list(
            map(
                lambda a: GitAutograderAnswersRecord.from_tuple(a),
                zip(self.questions, self.answers),
            )
        )

    def __getitem__(self, key: int) -> GitAutograderAnswersRecord:
        question = self.questions[key]
        answer = self.answers[key]
        return GitAutograderAnswersRecord.from_tuple((question, answer))

    def __len__(self) -> int:
        return len(self.questions)

    def get_by_question(self, question: str) -> Optional[GitAutograderAnswersRecord]:
        for i, q in enumerate(self.questions):
            if question == q:
                return GitAutograderAnswersRecord.from_tuple((q, self.answers[i]))
        return None



from typing import List, Optional, Tuple

from git import Commit, Repo
from git.diff import Lit_change_type

from git_autograder.diff import GitAutograderDiff, GitAutograderDiffHelper
from git_autograder.helpers.branch_helper import BranchHelper
from git_autograder.helpers.commit_helper import CommitHelper


class GraderHelper:
    def __init__(
        self,
        repo: Repo,
        branch_helper: BranchHelper,
        commit_helper: CommitHelper,
    ) -> None:
        self.repo = repo
        self.branch_helper = branch_helper
        self.commit_helper = commit_helper

    def has_non_empty_commits(self, branch: str = "main") -> bool:
        """Returns if a given branch has any non-empty commits."""
        for commit in self.commit_helper.user_commits(branch):
            if len(commit.stats.files) > 0:
                return True
        return False

    def has_edited_file(self, file_path: str, branch: str = "main") -> bool:
        """Returns if a given file has been edited in a given branch."""
        latest_commit = self.commit_helper.user_commits(branch)[-1]
        diff_helper = GitAutograderDiffHelper(
            self.commit_helper.start_commit(branch), latest_commit
        )
        for diff in diff_helper.iter_changes("M"):
            if diff.edited_file_path == file_path:
                return True
        return False

    def has_added_file(self, file_path: str, branch: str = "main") -> bool:
        """Returns if a given file has been added in a given branch."""
        latest_commit = self.commit_helper.user_commits(branch)[-1]
        diff_helper = GitAutograderDiffHelper(
            self.commit_helper.start_commit(branch), latest_commit
        )
        for diff in diff_helper.iter_changes("A"):
            if diff.edited_file_path == file_path:
                return True
        return False

    def get_file_diff(
        self, a: Commit, b: Commit, file_path: str
    ) -> Optional[Tuple[GitAutograderDiff, Lit_change_type]]:
        """Returns file difference between two commits across ALL change types."""
        # Based on the expectation that there can only exist one change type per file in a diff
        diff_helper = GitAutograderDiffHelper(a, b)
        change_types: List[Lit_change_type] = ["A", "D", "R", "M", "T"]
        for change_type in change_types:
            for change in diff_helper.iter_changes(change_type):
                if change.diff_parser is None or change.edited_file_path != file_path:
                    continue
                return change, change_type
        return None



import os
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import List, Optional

import pytz
from git import Repo

from git_autograder.answers.answers_parser import GitAutograderAnswersParser
from git_autograder.exception import (
    GitAutograderInvalidStateException,
    GitAutograderWrongAnswerException,
)
from git_autograder.helpers.answers_helper import AnswersHelper
from git_autograder.helpers.branch_helper import BranchHelper
from git_autograder.helpers.commit_helper import CommitHelper
from git_autograder.helpers.grader_helper import GraderHelper
from git_autograder.output import GitAutograderOutput
from git_autograder.status import GitAutograderStatus


class GitAutograderRepo:
    def __init__(
        self,
        is_local: bool,
        exercise_name: str,
        repo_path: Optional[str | os.PathLike] = None,
    ) -> None:
        # TODO: We should not be starting the grading at the point of initializing, but we're keeping this because of the exception system
        self.started_at = self.__now()
        self.is_local: bool = is_local
        self.exercise_name = exercise_name
        self.repo_path = (
            repo_path
            if repo_path is not None
            else Path.cwd().parent / "main"
            if not is_local
            else Path.cwd().parent / "exercises" / exercise_name
        )

        self.repo: Repo = Repo(self.repo_path)
        # Doing this to break the cyclic dependency
        self.branches: BranchHelper = BranchHelper(self.repo)
        self.commits: CommitHelper = CommitHelper(self.repo)
        self.grader: GraderHelper = GraderHelper(self.repo, self.branches, self.commits)
        self.__answers_parser: Optional[GitAutograderAnswersParser] = None
        self.__answers: Optional[AnswersHelper] = None

    @property
    def answers(self) -> AnswersHelper:
        """Parses a QnA file (answers.txt). Verifies that the file exists."""
        # We need to use singleton patterns here since we want to avoid repeatedly parsing
        # These are all optional to start since the grader might not require answers
        if self.__answers_parser is None:
            answers_file_path = Path(self.repo_path) / "answers.txt"
            # Use singleton for answers parser
            try:
                self.__answers_parser = GitAutograderAnswersParser(answers_file_path)
            except Exception as e:
                raise GitAutograderInvalidStateException(
                    str(e),
                )

        if self.__answers is None:
            self.__answers = AnswersHelper(self.repo, self.__answers_parser.answers)

        return self.__answers

    @staticmethod
    def __now() -> datetime:
        return datetime.now(tz=pytz.UTC)

    def to_output(
        self, comments: List[str], status: Optional[GitAutograderStatus] = None
    ) -> GitAutograderOutput:
        """
        Creates a GitAutograderOutput object.

        If there is no status provided, the status will be inferred from the comments.
        """
        return GitAutograderOutput(
            exercise_name=self.exercise_name,
            started_at=self.started_at,
            completed_at=self.__now(),
            is_local=self.is_local,
            comments=comments,
            status=(
                GitAutograderStatus.SUCCESSFUL
                if len(comments) == 0
                else GitAutograderStatus.UNSUCCESSFUL
            )
            if status is None
            else status,
        )

    def wrong_answer(self, comments: List[str]) -> GitAutograderWrongAnswerException:
        return GitAutograderWrongAnswerException(comments)



__all__ = [
    "AnswerRule",
    "HasExactValueRule",
    "NotEmptyRule",
    "HasExactListRule",
    "ContainsListRule",
]

from .answer_rule import AnswerRule
from .has_exact_value_rule import HasExactValueRule
from .not_empty_rule import NotEmptyRule
from .has_exact_list_rule import HasExactListRule
from .contains_list_rule import ContainsListRule



from dataclasses import dataclass
from typing import List, Tuple


@dataclass
class GitAutograderAnswersRecord:
    question: str
    answer: str

    def as_tuple(self) -> Tuple[str, str]:
        return self.question, self.answer

    @staticmethod
    def from_tuple(tuple_value: Tuple[str, str]) -> "GitAutograderAnswersRecord":
        return GitAutograderAnswersRecord(
            question=tuple_value[0], answer=tuple_value[1]
        )

    def answer_as_list(self) -> List[str]:
        points: List[str] = []
        acc = ""
        for line in self.answer.split("\n"):
            if line.startswith("-"):
                if acc.strip() != "":
                    points.append(acc.strip()[::])
                acc = line[1:].strip() + "\n"
            else:
                acc += line + "\n"
        if acc.strip() != "":
            points.append(acc.strip()[::])
        return points



from abc import ABC, abstractmethod

from git_autograder.answers.answers_record import GitAutograderAnswersRecord


class AnswerRule(ABC):
    @abstractmethod
    def apply(self, answer: GitAutograderAnswersRecord) -> None: ...



from typing import List

from git_autograder.answers.answers_record import GitAutograderAnswersRecord
from git_autograder.answers.rules.answer_rule import AnswerRule


class ContainsListRule(AnswerRule):
    def __init__(
        self, values: List[str], subset: bool = True, is_case_sensitive: bool = False
    ) -> None:
        self.values = values
        self.subset = subset
        self.is_case_sensitive = is_case_sensitive

    def apply(self, answer: GitAutograderAnswersRecord) -> None:
        expected = set(
            [v.lower() for v in self.values] if self.is_case_sensitive else self.values
        )
        given = (
            [v.lower() for v in answer.answer_as_list()]
            if self.is_case_sensitive
            else answer.answer_as_list()
        )
        if self.subset and not all([v in expected for v in given]):
            raise Exception(f"Answer for {answer.question} contains an invalid item.")
        elif not any([v in expected for v in given]):
            raise Exception(
                f"Answer for {answer.question} does not contain any valid items."
            )



from typing import List, Union


class GitAutograderException(Exception):
    def __init__(
        self,
        message: Union[str, List[str]],
    ) -> None:
        super().__init__(message)

        self.message = message


class GitAutograderInvalidStateException(GitAutograderException):
    def __init__(
        self,
        message: str,
    ) -> None:
        super().__init__(message)


class GitAutograderWrongAnswerException(GitAutograderException):
    def __init__(
        self,
        comments: List[str],
    ) -> None:
        super().__init__(comments)



from git_autograder.answers.answers_record import GitAutograderAnswersRecord
from git_autograder.answers.rules.answer_rule import AnswerRule


class ContainsValueRule(AnswerRule):
    def __init__(self, value: str, is_case_sensitive: bool = False) -> None:
        self.value = value
        self.is_case_sensitive = is_case_sensitive

    def apply(self, answer: GitAutograderAnswersRecord) -> None:
        expected = self.value.lower() if self.is_case_sensitive else self.value
        given = answer.answer.lower() if self.is_case_sensitive else answer.answer
        if given not in expected:
            raise Exception(
                f"Answer for {answer.question} does not contain the right answer."
            )



from typing import List

from git import Repo

from git_autograder.answers import GitAutograderAnswers
from git_autograder.answers.rules import AnswerRule
from git_autograder.exception import (
    GitAutograderInvalidStateException,
    GitAutograderWrongAnswerException,
)


class AnswersHelper:
    def __init__(self, repo: Repo, answers: GitAutograderAnswers) -> None:
        self.repo = repo
        self.answers = answers

    def validate_question(
        self, question: str, rules: List[AnswerRule]
    ) -> "AnswersHelper":
        answer = self.answers.get_by_question(question)
        if answer is None:
            raise GitAutograderInvalidStateException(
                f"Missing question {question} in answers file.",
            )

        for rule in rules:
            try:
                rule.apply(answer)
            except Exception as e:
                raise GitAutograderWrongAnswerException(
                    [str(e)],
                )

        return self



import os
from contextlib import contextmanager
from datetime import datetime
from typing import Callable, Iterator, Optional
from unittest import mock

import pytz
from git import Repo
from repo_smith.initialize_repo import RepoInitializer, initialize_repo

from git_autograder.exception import (
    GitAutograderInvalidStateException,
    GitAutograderWrongAnswerException,
)
from git_autograder.output import GitAutograderOutput
from git_autograder.repo import GitAutograderRepo
from git_autograder.status import GitAutograderStatus


def attach_start_tag(repo_initializer: RepoInitializer, step_id: str) -> None:
    def hook(r: Repo) -> None:
        all_commits = list(r.iter_commits())
        first_commit = list(reversed(all_commits))[0]
        first_commit_hash = first_commit.hexsha[:7]
        start_tag = f"git-mastery-start-{first_commit_hash}"
        r.create_tag(start_tag)

    repo_initializer.add_post_hook(step_id, hook)


def set_env(**kwargs) -> mock._patch_dict:
    return mock.patch.dict(os.environ, kwargs, clear=True)


@contextmanager
def setup_autograder(
    exercise_name: str,
    spec_path: str,
    step_id: str,
    grade_func: Callable[[GitAutograderRepo], GitAutograderOutput],
    setup: Callable[[Repo], None],
) -> Iterator[GitAutograderOutput]:
    repo_initializer = initialize_repo(spec_path)
    attach_start_tag(repo_initializer, step_id)
    with repo_initializer.initialize() as r:
        setup(r)
        output: Optional[GitAutograderOutput] = None
        started_at = datetime.now(tz=pytz.UTC)
        try:
            autograder = GitAutograderRepo(
                is_local=True, exercise_name=exercise_name, repo_path=r.working_dir
            )
            output = grade_func(autograder)
        except (
            GitAutograderInvalidStateException,
            GitAutograderWrongAnswerException,
        ) as e:
            output = GitAutograderOutput(
                exercise_name=exercise_name,
                started_at=started_at,
                completed_at=datetime.now(tz=pytz.UTC),
                is_local=True,
                comments=[e.message] if isinstance(e.message, str) else e.message,
                status=(
                    GitAutograderStatus.ERROR
                    if isinstance(e, GitAutograderInvalidStateException)
                    else GitAutograderStatus.UNSUCCESSFUL
                ),
            )
        except Exception as e:
            # Unexpected exception
            output = GitAutograderOutput(
                exercise_name=exercise_name,
                started_at=None,
                completed_at=None,
                is_local=True,
                comments=[str(e)],
                status=GitAutograderStatus.ERROR,
            )

        assert output is not None
        yield output



__all__ = [
    "autograder",
    "setup_autograder",
    "set_env",
    "GitAutograderRepo",
    "GitAutograderStatus",
    "GitAutograderOutput",
]

from .autograder import autograder
from .test_utils import setup_autograder, set_env
from .repo import GitAutograderRepo
from .status import GitAutograderStatus
from .output import GitAutograderOutput
